require "Apollo" 
require "ApolloTimer"
require "GameLib"
require "XmlDoc"

local AlpacaDebug = {} 

function AlpacaDebug:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaDebug:Init()
  local bHasConfigureFunction = false
  local strConfigureButtonText = ""
  local tDependencies = {}
	
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function AlpacaDebug:OnLoad()
  self.tCounters = {}
  
	self.xmlDoc = XmlDoc.CreateFromFile("AlpacaDebug.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

function AlpacaDebug:OnDocLoaded()
  if self.xmlDoc == nil or not self.xmlDoc:IsLoaded() then return end

  self.wndForm = Apollo.LoadForm(self.xmlDoc, "Form", nil, self)
  self.wndForm:Show(false, true)
  self.wndForm:SetSizingMinimum(300, 200)

  Apollo.RegisterEventHandler("AlpacaOpenDebug", "OnAlpacaOpenDebug", self)
  Apollo.RegisterEventHandler("AlpacaSetDebugCounter", "OnAlpacaSetDebugCounter", self)
end

-- Events

function AlpacaDebug:OnAlpacaOpenDebug()
  self:OpenForm()
end

function AlpacaDebug:OnAlpacaSetDebugCounter(strName, fValue)
  self.tCounters[strName] = fValue
end

function AlpacaDebug:OnWindowClosed(wndHandler, wndControl)
  if wndHandler == wndControl then
    self:CloseForm()	
  end
end

-- Buttons

function AlpacaDebug:OnCloseButtonSignal(wndHandler, wndControl, eMouseButton)
  if wndHandler == wndControl then
    self:CloseForm()	
  end
end

-- Timers

function AlpacaDebug:OnTimer()
	if not self.wndForm:IsShown() then return end

  -- update buffs
	local wndBuffs = self.wndForm:FindChild("Buffs")
  
	local fnGetBuffs = function()
		if not GameLib.GetPlayerUnit() then return {} end
		return GameLib.GetPlayerUnit():GetBuffs().arBeneficial	
	end

	for _, tBuff in ipairs(fnGetBuffs()) do
		local strName = tBuff.splEffect:GetName()
		if not self.arBuffsSeen[strName] then
			self.arBuffsSeen[strName] = true
			wndBuffs:SetText(wndBuffs:GetText() .. strName .. "\n")
		end
	end
  
  -- update counters
  local wndCounters = self.wndForm:FindChild("Counters")
  
  wndCounters:SetText("")
  
  local tAddonInfo = Apollo.GetAddonInfo("AlpacaHUD")
  
  wndCounters:SetText(wndCounters:GetText() .. "fCallTimePerFrame: " .. (tAddonInfo.fCallTimePerFrame * 1000) .. "\n")
  wndCounters:SetText(wndCounters:GetText() .. "fCallTimePerSecond: " .. (tAddonInfo.fCallTimePerSecond * 1000) .. "\n")  
  wndCounters:SetText(wndCounters:GetText() .. "nMemoryUsage: " .. (tAddonInfo.nMemoryUsage / 1000) .. "\n")
  
  local arCounterKeys = {}
  
  for strKey, _ in pairs(self.tCounters) do
    table.insert(arCounterKeys, strKey)
  end
  
  table.sort(arCounterKeys)
  
  for _, strKey in ipairs(arCounterKeys) do
    wndCounters:SetText(wndCounters:GetText() .. strKey .. ": " .. self.tCounters[strKey] .. "\n")
  end
end

-- Controls

function AlpacaDebug:OpenForm()
	local wndForm = self.wndForm
	local wndBuffs = self.wndForm:FindChild("Buffs")

	self.arBuffsSeen = {}
	wndBuffs:SetText("")

	wndForm:Show(true)
	self.timer = ApolloTimer.Create(1, true, "OnTimer", self)
end

function AlpacaDebug:CloseForm()
	self.wndForm:Show(false)
	self.timer = nil
end

-- Start

local AlpacaDebugInst = AlpacaDebug:new()
AlpacaDebugInst:Init()
