require "Apollo"
require "ApolloColor"
require "GameLib"

local AlpacaGameLibrary = nil
local AlpacaPluginLibrary = nil
local AlpacaStandardLibrary = nil
local AlpacaWindowLibrary = nil

local AlpacaBars = {}

-- Plugin

function AlpacaBars:PluginId()
  return "AlpacaBars"
end

function AlpacaBars:PluginName()
  return "Bars"
end

function AlpacaBars:PluginDefaults() 
  local fnCreateColor = function(...)
    return AlpacaWindowLibrary:AlpacaColor(...)
  end
 
  local fnCreateResourceColors = function(fRed, fGreen, fBlue, fAlpha)
    return {
      crBorder = fnCreateColor(fRed, fGreen, fBlue, fAlpha),
      crFill = fnCreateColor(fRed, fGreen, fBlue, fAlpha),
      crMarker = fnCreateColor(fRed, fGreen, fBlue, fAlpha)
    }
  end

  local tDefaults = {
    bEnable = true,
    bEnableCarbinePlayerFrame = true,
    bEnableCarbineTargetFrame = true,
    tHealthBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 80, nVertical = 375 },
      tSize = { nWidth = 15, nHeight = 300 },
      crBorder = fnCreateColor(0.14, 0.85, 0.14, 0.5),
      crFill = fnCreateColor(0.14, 0.85, 0.14, 0.5)
    },
    tShieldBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 60, nVertical = 400 },
      tSize = { nWidth = 15, nHeight = 250 },
      crBorder = fnCreateColor(0.33, 0.85, 0.73, 0.5),
      crFill = fnCreateColor(0.33, 0.85, 0.73, 0.5)
    },
    tAbsorbBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = false,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 40, nVertical = 425 },
      tSize = { nWidth = 15, nHeight = 200 },
      crBorder = fnCreateColor(1, 0.5, 0, 0.5),
      crFill = fnCreateColor(1, 0.5, 0, 0.5)
    },
    tTargetHealthBar = {
      bEnable = false,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 815, nVertical = 425 },
      tSize = { nWidth = 15, nHeight = 200 },
      crBorder = fnCreateColor(0.14, 0.85, 0.14, 0.5),
      crFill = fnCreateColor(0.14, 0.85, 0.14, 0.5)
    },
    tTargetShieldBar = {
      bEnable = false,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 835, nVertical = 450 },
      tSize = { nWidth = 15, nHeight = 150 },
      crBorder = fnCreateColor(0.33, 0.85, 0.73, 0.5),
      crFill = fnCreateColor(0.33, 0.85, 0.73, 0.5)
    },
    tTargetAbsorbBar = {
      bEnable = false,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = false,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 855, nVertical = 475 },
      tSize = { nWidth = 15, nHeight = 100 },
      crBorder = fnCreateColor(1, 0.5, 0, 0.5),
      crFill = fnCreateColor(1, 0.5, 0, 0.5)
    },
    tCastBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = false,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Right,
      tPosition = { nHorizontal = 500, nVertical = 570 },
      tSize = { nWidth = 200, nHeight = 10 },
      crBorder = fnCreateColor(1, 1, 0, 0.5),
      crFill = fnCreateColor(1, 1, 0, 0.5)
    },
    tDashBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = false,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Right,
      tPosition = { nHorizontal = 500, nVertical = 585 },
      tSize = { nWidth = 200, nHeight = 10 },
      crBorder = fnCreateColor(1, 1, 1, 0.5),
      crFill = fnCreateColor(1, 1, 1, 0.5),
      tMarker = { 
        nWidth = 2, 
        nHeight = 10,
        crColor = fnCreateColor(1, 1, 1, 0.5)
      },
    },
    tSprintBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = false,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Right,
      tPosition = { nHorizontal = 500, nVertical = 600 },
      tSize = { nWidth = 200, nHeight = 10 },
      crBorder = fnCreateColor(1, 1, 1, 0.5),
      crFill = fnCreateColor(1, 1, 1, 0.5)
    },
    tFocusBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = false,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 100, nVertical = 350 },
      tSize = { nWidth = 10, nHeight = 350 },
      crBorder = fnCreateColor(0, 0, 1, 0.5),
      crFill = fnCreateColor(0, 0, 1, 0.5)
    },
    tResourceBar = {
      bEnable = true,
      bOnlyShowInCombat = false,
      bShowWhenEmpty = true,
      bShowWhenFull = true,
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tPosition = { nHorizontal = 115, nVertical = 350 },
      tSize = { nWidth = 20, nHeight = 350 },
      tMarker = { 
        nWidth = 20, 
        nHeight = 2,
      },
      tClassColors = {
        [GameLib.CodeEnumClass.Engineer]      = fnCreateResourceColors(1, 1, 1, 0.5),
        [GameLib.CodeEnumClass.Esper]         = fnCreateResourceColors(1, 1, 1, 0.5),
        [GameLib.CodeEnumClass.Medic]         = fnCreateResourceColors(1, 1, 1, 0.5),
        [GameLib.CodeEnumClass.Spellslinger]  = fnCreateResourceColors(1, 1, 1, 0.5),
        [GameLib.CodeEnumClass.Stalker]       = fnCreateResourceColors(1, 1, 1, 0.5),
        [GameLib.CodeEnumClass.Warrior]       = fnCreateResourceColors(1, 1, 1, 0.5)
      },
      tClassMarkers = {
        [GameLib.CodeEnumClass.Engineer]      = { 0.3,  0.7,  0,    0,  },
        [GameLib.CodeEnumClass.Esper]         = { 0.2,  0.3,  0.4,  0.5 },
        [GameLib.CodeEnumClass.Medic]         = { 0.25, 0.5,  0.75, 0   },
        [GameLib.CodeEnumClass.Spellslinger]  = { 0.25, 0.5,  0.75, 0,  },
        [GameLib.CodeEnumClass.Stalker]       = { 0,    0,    0,    0,  },
        [GameLib.CodeEnumClass.Warrior]       = { 0.25, 0.5,  0.75, 0,  }
      }     
    }
  }  

  return tDefaults
end

function AlpacaBars:PluginForm() 
  local fnCreateDirectionOptions = function()
    return {
      { strLabel = "Left", nValue = AlpacaWindowLibrary.EnumDirection.Left },
      { strLabel = "Right", nValue = AlpacaWindowLibrary.EnumDirection.Right },
      { strLabel = "Up", nValue = AlpacaWindowLibrary.EnumDirection.Up },
      { strLabel = "Down", nValue = AlpacaWindowLibrary.EnumDirection.Down }
    }
  end
  
  local arInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Bars"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                      { "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Enable Carbine Player Frame", { "bEnableCarbinePlayerFrame" }),
    AlpacaPluginLibrary:CreateCheckbox("Enable Carbine Target Frame", { "bEnableCarbineTargetFrame" }),
  }
   
  local arPlayerInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Health Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tHealthBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tHealthBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tHealthBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tHealthBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tHealthBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tHealthBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tHealthBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tHealthBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tHealthBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tHealthBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tHealthBar", "crBorder" }),
    
    AlpacaPluginLibrary:CreateHeader("Shield Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tShieldBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tShieldBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tShieldBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tShieldBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tShieldBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tShieldBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tShieldBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tShieldBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tShieldBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tShieldBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tShieldBar", "crBorder" }),
    
    AlpacaPluginLibrary:CreateHeader("Absorb Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tAbsorbBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tAbsorbBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tAbsorbBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tAbsorbBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tAbsorbBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tAbsorbBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tAbsorbBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tAbsorbBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tAbsorbBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tAbsorbBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tAbsorbBar", "crBorder" }),
      
    AlpacaPluginLibrary:CreateHeader("Cast Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tCastBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tCastBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tCastBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tCastBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tCastBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tCastBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tCastBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tCastBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tCastBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tCastBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tCastBar", "crBorder" }),
    
    AlpacaPluginLibrary:CreateHeader("Dash Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tDashBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tDashBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tDashBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tDashBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tDashBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tDashBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tDashBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tDashBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Marker (Width)",          { "tDashBar", "tMarker", "nWidth" }, 0, 100, 1),
    AlpacaPluginLibrary:CreateSlider("Marker (Height)",         { "tDashBar", "tMarker", "nHeight" }, 0, 100, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tDashBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tDashBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tDashBar", "crBorder" }),
    AlpacaPluginLibrary:CreateColor("Marker Color",             { "tDashBar", "tMarker", "crColor" }),
    
    AlpacaPluginLibrary:CreateHeader("Sprint Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tSprintBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tSprintBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tSprintBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tSprintBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tSprintBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tSprintBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tSprintBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tSprintBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tSprintBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tSprintBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tSprintBar", "crBorder" })
  }
  
  local arResourceInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Focus Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tFocusBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tFocusBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tFocusBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tFocusBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tFocusBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tFocusBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tFocusBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tFocusBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tFocusBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tFocusBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tFocusBar", "crBorder" }),
    
    AlpacaPluginLibrary:CreateHeader("Resource Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tResourceBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tResourceBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tResourceBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tResourceBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tResourceBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tResourceBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tResourceBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tResourceBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Marker (Width)",          { "tResourceBar", "tMarker", "nWidth" }, 0, 100, 1),
    AlpacaPluginLibrary:CreateSlider("Marker (Height)",         { "tResourceBar", "tMarker", "nHeight" }, 0, 100, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tResourceBar", "eFillDirection" }, fnCreateDirectionOptions())
  }
   
  local tClasses = {
    [GameLib.CodeEnumClass.Engineer] = "Engineer",
    [GameLib.CodeEnumClass.Esper] = "Esper",
    [GameLib.CodeEnumClass.Medic] = "Medic",
    [GameLib.CodeEnumClass.Spellslinger] = "Spellslinger",
    [GameLib.CodeEnumClass.Stalker] = "Stalker",
    [GameLib.CodeEnumClass.Warrior] = "Warrior"
  }
  
  local fnCreateClassResourceSubSection = function(nClassId, strClassName)
    return {
      AlpacaPluginLibrary:CreateSubHeader(strClassName),
      AlpacaPluginLibrary:CreateColor("Bar Color",                { "tResourceBar", "tClassColors", nClassId, "crFill" }),
      AlpacaPluginLibrary:CreateColor("Border Color",             { "tResourceBar", "tClassColors", nClassId, "crBorder" }),
      AlpacaPluginLibrary:CreateColor("Marker Color",             { "tResourceBar", "tClassColors", nClassId, "crMarker" }),
      AlpacaPluginLibrary:CreateSlider("Marker 1 Percent",        { "tResourceBar", "tClassMarkers", nClassId, 1 }, 0, 1, 0.01),
      AlpacaPluginLibrary:CreateSlider("Marker 2 Percent",        { "tResourceBar", "tClassMarkers", nClassId, 2 }, 0, 1, 0.01),
      AlpacaPluginLibrary:CreateSlider("Marker 3 Percent",        { "tResourceBar", "tClassMarkers", nClassId, 3 }, 0, 1, 0.01),
      AlpacaPluginLibrary:CreateSlider("Marker 4 Percent",        { "tResourceBar", "tClassMarkers", nClassId, 4 }, 0, 1, 0.01)
    }
  end
  
  for nClassId, strClassName in pairs(tClasses) do
    for _, tInput in ipairs(fnCreateClassResourceSubSection(nClassId, strClassName)) do
      table.insert(arResourceInputDefinitions, tInput)
    end
  end
  
  local arTargetInputDefinitions = { 
    AlpacaPluginLibrary:CreateHeader("Target Health Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tTargetHealthBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tTargetHealthBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tTargetHealthBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tTargetHealthBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tTargetHealthBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tTargetHealthBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tTargetHealthBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tTargetHealthBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tTargetHealthBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tTargetHealthBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tTargetHealthBar", "crBorder" }),
    
    AlpacaPluginLibrary:CreateHeader("Target Shield Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tTargetShieldBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tTargetShieldBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tTargetShieldBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tTargetShieldBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tTargetShieldBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tTargetShieldBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tTargetShieldBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tTargetShieldBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tTargetShieldBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tTargetShieldBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tTargetShieldBar", "crBorder" }),
    
    AlpacaPluginLibrary:CreateHeader("Target Absorb Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "tTargetAbsorbBar", "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tTargetAbsorbBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Empty",       { "tTargetAbsorbBar", "bShowWhenEmpty" }),
    AlpacaPluginLibrary:CreateCheckbox("Show When Full",        { "tTargetAbsorbBar", "bShowWhenFull" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tTargetAbsorbBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tTargetAbsorbBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "tTargetAbsorbBar", "tSize", "nWidth" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "tTargetAbsorbBar", "tSize", "nHeight" }, 10, 2000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",          { "tTargetAbsorbBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tTargetAbsorbBar", "crFill" }),
    AlpacaPluginLibrary:CreateColor("Border Color",             { "tTargetAbsorbBar", "crBorder" })
  }
  
  local arSubForms = {
     AlpacaPluginLibrary:CreateSubForm("Player", arPlayerInputDefinitions),
     AlpacaPluginLibrary:CreateSubForm("Resource", arResourceInputDefinitions),
     AlpacaPluginLibrary:CreateSubForm("Target", arTargetInputDefinitions)
  }
  
  return AlpacaPluginLibrary:CreateForm(arInputDefinitions, arSubForms)
end

function AlpacaBars:OnPluginRegistered(wndDisplay)
  self.wndDisplay = wndDisplay
end

function AlpacaBars:OnPluginSettingsChanged(tSettings)
  self.tSettings = tSettings
  self:RequestRebuild(0)
end

-- Addon

function AlpacaBars:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaBars:Init()
  local bHasConfigureFunction = false
  local strConfigureButtonText = nil
  local tDependencies = { "AlpacaStandardLibrary", "AlpacaPluginLibrary", "AlpacaGameLibrary", "AlpacaWindowLibrary" }

  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function AlpacaBars:OnLoad() 
  Apollo.LoadSprites("AlpacaSprites.xml")
  
  AlpacaGameLibrary = Apollo.GetPackage("AlpacaGameLibrary").tPackage
  AlpacaPluginLibrary = Apollo.GetPackage("AlpacaPluginLibrary").tPackage
  AlpacaStandardLibrary = Apollo.GetPackage("AlpacaStandardLibrary").tPackage
  AlpacaWindowLibrary = Apollo.GetPackage("AlpacaWindowLibrary").tPackage
  
 	Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)
  
  AlpacaPluginLibrary:RegisterPlugin(self)
end

-- Timers

function AlpacaBars:RequestRebuild(fDelay)
  self.fRebuildTime = os.clock() + (fDelay or 0)
end

function AlpacaBars:IsReady() 
  if not self.bReady then
    if not AlpacaWindowLibrary:IsLoaded() then 
      return false
    end
    
    if not self.wndDisplay then
      return false
    end
    
    if not self.tSettings then
      return false
    end

    self.bReady = true
  end
  
  return self.bReady
end

function AlpacaBars:OnFrame()  
  if not self:IsReady() then
    return
  end
  
  self.unitPlayer = AlpacaGameLibrary:GetPlayerUnit()

  if not self.unitPlayer then
    return
  end

  self:InterceptCarbineSprintMeterEvents()
  self:InterceptCarbineUnitFramesEvents()

  local bRebuild = self.fRebuildTime and os.clock() >= self.fRebuildTime 

  if not self.wndHealthBar or bRebuild then
    self:CreateHealthBar()
  end  

  if not self.wndShieldBar or bRebuild then
    self:CreateShieldBar()
  end
    
  if not self.wndAbsorbBar or bRebuild then
    self:CreateAbsorbBar()
  end
  
  if not self.wndTargetHealthBar or bRebuild then
    self:CreateTargetHealthBar()
  end  

  if not self.wndTargetShieldBar or bRebuild then
    self:CreateTargetShieldBar()
  end
    
  if not self.wndTargetAbsorbBar or bRebuild then
    self:CreateTargetAbsorbBar()
  end
  if not self.wndCastBar or bRebuild then
    self:CreateCastBar()
  end

  if not self.wndSprintBar or bRebuild then
    self:CreateSprintBar()
  end
  
  if not self.wndDashBar or bRebuild then
    self:CreateDashBar()
  end

  if not self.wndFocusBar or bRebuild then
    self:CreateFocusBar()
  end

  if not self.wndResourceBar or bRebuild then
    self:CreateResourceBar()
  end

  if (bRebuild) then
    self.fRebuildTime = nil
  end

  self:RefreshHealthBar()
  self:RefreshShieldBar()
  self:RefreshAbsorbBar()
  self:RefreshTargetHealthBar()
  self:RefreshTargetShieldBar()
  self:RefreshTargetAbsorbBar()
  self:RefreshCastBar()
  self:RefreshSprintBar()
  self:RefreshDashBar()
  self:RefreshFocusBar()
  self:RefreshResourceBar()
end

-- HealthBar

function AlpacaBars:CreateHealthBar()
  if self.wndHealthBar then 
    self.wndHealthBar:Destroy() 
    self.wndHealthBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tHealthBar.bEnable then return end

  self.wndHealthBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tHealthBar, self.wndDisplay)
end

function AlpacaBars:RefreshHealthBar()
  if not self.wndHealthBar then return end

  local fPercent = AlpacaGameLibrary:GetUnitHealthPercent(self.unitPlayer)

  AlpacaWindowLibrary:UpdateBarFill(self.wndHealthBar, self.tSettings.tHealthBar, fPercent)
  
  self:ToggleBar(self.wndHealthBar, self.tSettings.tHealthBar, fPercent)
end

-- ShieldBar

function AlpacaBars:CreateShieldBar()
  if self.wndShieldBar then 
    self.wndShieldBar:Destroy() 
    self.wndShieldBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tShieldBar.bEnable then return end

  self.wndShieldBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tShieldBar, self.wndDisplay)
end

function AlpacaBars:RefreshShieldBar()
  if not self.wndShieldBar then return end
 
  local fPercent = AlpacaGameLibrary:GetUnitShieldPercent(self.unitPlayer)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndShieldBar, self.tSettings.tShieldBar, fPercent)
  
  self:ToggleBar(self.wndShieldBar, self.tSettings.tShieldBar, fPercent)
end

-- AbsorbBar

function AlpacaBars:CreateAbsorbBar()
  if self.wndAbsorbBar then 
    self.wndAbsorbBar:Destroy() 
    self.wndAbsorbBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tAbsorbBar.bEnable then return end

  self.wndAbsorbBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tAbsorbBar, self.wndDisplay)
end

function AlpacaBars:RefreshAbsorbBar()
  if not self.wndAbsorbBar then return end

  local fPercent = AlpacaGameLibrary:GetUnitAbsorbPercent(self.unitPlayer)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndAbsorbBar, self.tSettings.tAbsorbBar, fPercent)
  
  self:ToggleBar(self.wndAbsorbBar, self.tSettings.tAbsorbBar, fPercent)
end

-- TargetHealthBar

function AlpacaBars:CreateTargetHealthBar()
  if self.wndTargetHealthBar then 
    self.wndTargetHealthBar:Destroy() 
    self.wndTargetHealthBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tTargetHealthBar.bEnable then return end

  self.wndTargetHealthBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tTargetHealthBar, self.wndDisplay)
end

function AlpacaBars:RefreshTargetHealthBar()
  if not self.wndTargetHealthBar then return end

  local unitTarget = GameLib.GetTargetUnit()

  if not unitTarget then
    self.wndTargetHealthBar:Show(self.tSettings.bPreview, true)
    return
  end

  local fPercent = AlpacaGameLibrary:GetUnitHealthPercent(unitTarget)

  AlpacaWindowLibrary:UpdateBarFill(self.wndTargetHealthBar, self.tSettings.tTargetHealthBar, fPercent)
  
  self:ToggleBar(self.wndTargetHealthBar, self.tSettings.tTargetHealthBar, fPercent)
end

-- TargetShieldBar

function AlpacaBars:CreateTargetShieldBar()
  if self.wndTargetShieldBar then 
    self.wndTargetShieldBar:Destroy() 
    self.wndTargetShieldBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tTargetShieldBar.bEnable then return end

  self.wndTargetShieldBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tTargetShieldBar, self.wndDisplay)
end

function AlpacaBars:RefreshTargetShieldBar()
  if not self.wndTargetShieldBar then return end
 
  local unitTarget = GameLib.GetTargetUnit()

  if not unitTarget then
    self.wndTargetShieldBar:Show(self.tSettings.bPreview, true)
    return
  end
 
  local fPercent = AlpacaGameLibrary:GetUnitShieldPercent(unitTarget)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndTargetShieldBar, self.tSettings.tTargetShieldBar, fPercent)
  
  self:ToggleBar(self.wndTargetShieldBar, self.tSettings.tTargetShieldBar, fPercent)
end

-- TargetAbsorbBar

function AlpacaBars:CreateTargetAbsorbBar()
  if self.wndTargetAbsorbBar then 
    self.wndTargetAbsorbBar:Destroy() 
    self.wndTargetAbsorbBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tTargetAbsorbBar.bEnable then return end

  self.wndTargetAbsorbBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tTargetAbsorbBar, self.wndDisplay)
end

function AlpacaBars:RefreshTargetAbsorbBar()
  if not self.wndTargetAbsorbBar then return end

  local unitTarget = GameLib.GetTargetUnit()

  if not unitTarget then
    self.wndTargetAbsorbBar:Show(self.tSettings.bPreview, true)
    return
  end

  local fPercent = AlpacaGameLibrary:GetUnitAbsorbPercent(unitTarget)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndTargetAbsorbBar, self.tSettings.tTargetAbsorbBar, fPercent)
  
  self:ToggleBar(self.wndTargetAbsorbBar, self.tSettings.tTargetAbsorbBar, fPercent)
end

-- CastBar

function AlpacaBars:CreateCastBar()
  if self.wndCastBar then 
    self.wndCastBar:Destroy() 
    self.wndCastBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tCastBar.bEnable then return end

  self.wndCastBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tCastBar, self.wndDisplay)
end

function AlpacaBars:RefreshCastBar()
  if not self.wndCastBar then return end

  local fPercent = AlpacaGameLibrary:GetUnitCastPercent(self.unitPlayer)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndCastBar, self.tSettings.tCastBar, fPercent)
  
  self:ToggleBar(self.wndCastBar, self.tSettings.tCastBar, fPercent)
end

-- SprintBar

function AlpacaBars:CreateSprintBar()
  if self.wndSprintBar then 
    self.wndSprintBar:Destroy() 
    self.wndSprintBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tSprintBar.bEnable then
    return
  end
  
  self.wndSprintBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tSprintBar, self.wndDisplay)
end

function AlpacaBars:RefreshSprintBar()
  if not self.wndSprintBar then return end

  local fPercent = AlpacaGameLibrary:GetUnitSprintPercent(self.unitPlayer)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndSprintBar, self.tSettings.tSprintBar, fPercent)
  
  self:ToggleBar(self.wndSprintBar, self.tSettings.tSprintBar, fPercent)
end

-- DashBar

function AlpacaBars:CreateDashBar()
  if self.wndDashBar then 
    self.wndDashBar:Destroy() 
    self.wndDashBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tDashBar.bEnable then return end

  self.wndDashBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tDashBar, self.wndDisplay)
  
  AlpacaWindowLibrary:CreateMarker(AlpacaWindowLibrary.EnumMarkerType.Line, self.tSettings.tDashBar, self.wndDashBar, 0.5)
  
end

function AlpacaBars:RefreshDashBar()
  if not self.wndDashBar then return end

  local fPercent = AlpacaGameLibrary:GetUnitDashPercent(self.unitPlayer)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndDashBar, self.tSettings.tDashBar, fPercent)
  
  self:ToggleBar(self.wndDashBar, self.tSettings.tDashBar, fPercent)
end

-- FocusBar

function AlpacaBars:CreateFocusBar()
  if self.wndFocusBar then 
    self.wndFocusBar:Destroy() 
    self.wndFocusBar = nil
  end
  
  if not self.tSettings.bEnable or not self.tSettings.tFocusBar.bEnable then return end

  self.wndFocusBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tFocusBar, self.wndDisplay)
end

function AlpacaBars:RefreshFocusBar()
  if not self.wndFocusBar then return end

  local fPercent = AlpacaGameLibrary:GetUnitFocusPercent(self.unitPlayer)
  
  AlpacaWindowLibrary:UpdateBarFill(self.wndFocusBar, self.tSettings.tFocusBar, fPercent)
  
  self:ToggleBar(self.wndFocusBar, self.tSettings.tFocusBar, fPercent)
end

-- ResourceBar

function AlpacaBars:CreateResourceBar()
  if self.wndResourceBar then 
    self.wndResourceBar:Destroy() 
    self.wndResourceBar = nil
  end
    
  if not self.tSettings.bEnable or not self.tSettings.tResourceBar.bEnable then return end
  
  local wndBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tResourceBar, self.wndDisplay)
  local wndProgressBar = wndBar:FindChild("ProgressBar")
  local nClassId = self.unitPlayer:GetClassId()  
  
  for _, fPercent in ipairs(self.tSettings.tResourceBar.tClassMarkers[nClassId]) do
    if fPercent > 0 and fPercent < 1 then
      AlpacaWindowLibrary:CreateMarker(AlpacaWindowLibrary.EnumMarkerType.Line, self.tSettings.tResourceBar, wndProgressBar, fPercent)
    end
  end
  
  local crBorder = self.tSettings.tResourceBar.tClassColors[nClassId].crBorder
  local crFill = self.tSettings.tResourceBar.tClassColors[nClassId].crFill
  local crMarker = self.tSettings.tResourceBar.tClassColors[nClassId].crMarker

  local wndBorder = wndBar:FindChild("Border")
  wndBorder:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crBorder))

  local wndFill = wndBar:FindChild("Fill")  
  wndFill:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crFill))
  
  for _, wndFrame in ipairs(wndProgressBar:GetChildren()) do
    if wndFrame:GetName() == "Marker" then
      wndFrame:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crMarker))
    end
  end
  
  self.wndResourceBar = wndBar
end

function AlpacaBars:RefreshResourceBar()
  if not self.wndResourceBar then return end
  
  local nClassId = self.unitPlayer:GetClassId()
  local fPercent = AlpacaGameLibrary:GetUnitResourcePercent(self.unitPlayer)
           
  if nClassId == GameLib.CodeEnumClass.Medic then
    local tBuff = AlpacaGameLibrary:GetBuff(self.unitPlayer, "Power Charge")
    local nPowerCharges = tBuff and tBuff.nCount or 0

    fPercent = math.min(fPercent + nPowerCharges * .0833, 1)
    
    -- use cached value after a reduction in powercharges (removes flicker after 3rd charge)
    if nPowerCharges >= (self.nMedicPowerCharges or 0) then
      self.nMedicPowerCharges = nPowerCharges
      self.fMedicResourcePercent = fPercent
      self.fMedicResourcePercentCacheTime = os.clock()
    else
      if os.clock() - self.fMedicResourcePercentCacheTime < 0.2 then
        fPercent = self.fMedicResourcePercent 
      else
        self.nMedicPowerCharges = nil
        self.fMedicResourcePercent = nil
        self.fMedicResourcePercentCacheTime = nil
      end
    end
  end

  AlpacaWindowLibrary:UpdateBarFill(self.wndResourceBar, self.tSettings.tResourceBar, fPercent)
  
  self:ToggleBar(self.wndResourceBar, self.tSettings.tResourceBar, fPercent)
end

-- Helpers

function AlpacaBars:ToggleBar(wndBar, tBarSettings, fPercent)   
  local bShow = true

  if not tBarSettings.bShowWhenEmpty and fPercent <= 0 then
    bShow = false
  end

  if not tBarSettings.bShowWhenFull and fPercent >= 1 then
    bShow = false
  end

  if tBarSettings.bOnlyShowInCombat and not AlpacaGameLibrary:IsInCombatOrDamaged(self.unitPlayer) then
    bShow = false
  end
  
  if self.tSettings.bPreview then
    bShow = true
  end

  wndBar:Show(bShow, true)
end

-- Carbine Addons

function AlpacaBars:InterceptCarbineSprintMeterEvents() 
  local objCarbineAddon = Apollo.GetAddon("SprintMeter")
  
  if not objCarbineAddon then
    return
  end
  
  if not self.fnSprintMeterOnFrameBackup then
    self.fnSprintMeterOnFrameBackup = objCarbineAddon.OnFrame
  end
  
  if self.tSettings.tSprintBar.bEnable then
    objCarbineAddon.OnFrame = function ()      
      objCarbineAddon.wndMain:Show(false, true)
    end    
  else
    objCarbineAddon.OnFrame = self.fnSprintMeterOnFrameBackup
  end  
end

function AlpacaBars:InterceptCarbineUnitFramesEvents()
  if self.bInterceptCarbineUnitFramesEvents then
    return
  end
  
  local objCarbineAddon = Apollo.GetAddon("TargetFrame")
    
  if not objCarbineAddon then
    return
  end
  
  Apollo.RemoveEventHandler("TargetUnitChanged", objCarbineAddon)
  
  objCarbineAddon.OnCharacterLoaded = function ()
    local unitPlayer = GameLib.GetPlayerUnit()
	
    if not unitPlayer then
      return
    end
      
    objCarbineAddon.luaUnitFrame:SetTarget(self.tSettings.bEnableCarbinePlayerFrame and unitPlayer or nil)
    objCarbineAddon.luaTargetFrame:SetTarget(self.tSettings.bEnableCarbineTargetFrame and unitPlayer:GetTarget() or nil)
    objCarbineAddon.luaFocusFrame:SetTarget(unitPlayer:GetAlternateTarget())
  end

  objCarbineAddon.OnTargetUnitChanged = function (unitTarget)
    objCarbineAddon.luaTargetFrame:SetTarget(self.tSettings.bEnableCarbineTargetFrame and unitTarget)
  end
  
  self.bInterceptCarbineUnitFramesEvents = true
end

-- Start

local AlpacaBarsInst = AlpacaBars:new()
AlpacaBarsInst:Init()
