require "Apollo"
require "GameLib"


local AlpacaGameLibrary = nil
local AlpacaPluginLibrary = nil
local AlpacaStandardLibrary = nil
local AlpacaWindowLibrary = nil

local AlpacaProcs = {} 

-- Plugin

function AlpacaProcs:PluginId()
  return "AlpacaProcs"
end

function AlpacaProcs:PluginName()
  return "Procs"
end

function AlpacaProcs:PluginDefaults() 
  local fnCreateColor = function(...)
    return AlpacaWindowLibrary:AlpacaColor(...)
  end
 
  local tDefaults = {
    bEnable = true,
    bOnlyShowInCombat = false,
    tPosition = { nHorizontal = 900, nVertical = 400 },
    nSpacing = 40,
    tProcBar = {
      eFillDirection = AlpacaWindowLibrary.EnumDirection.Up,
      tSize = { nWidth = 20, nHeight = 250 },
      tMarker = { 
        nWidth = 40, 
        nHeight = 20,
        crColor = fnCreateColor(1, 0.5, 0)
      },
      crCooldown = fnCreateColor(0.8, 0.8, 0.8),
      nIconSize = 20,
      eIconPosition = AlpacaWindowLibrary.EnumPosition.Bottom,
      tAbilityColors = {
        -- Engineer
        ["Quick Burst"]        = fnCreateColor(1, 0, 0, 0.7),
        ["Feedback"]           = fnCreateColor(0, 1, 0, 0.7),
        -- Medic
        ["Atomize"]            = fnCreateColor(1, 0.63, 0.21, 0.7),
        ["Dual Shock"]         = fnCreateColor(0, 1, 0.7),
        -- Spellslinger
        ["Flame Burst"]        = fnCreateColor(1, 0.85, 0.04, 0.7),
        -- Stalker
        ["Punish"]             = fnCreateColor(0, 0.90, 0.90, 0.7),
        ["Decimate"]           = fnCreateColor(0.92, 0.74, 0, 0.7),
        -- Warrior
        ["Breaching Strikes"]  = fnCreateColor(1, 0, 0, 0.7),
        ["Atomic Spear"]       = fnCreateColor(1, 0.85, 0.04, 0.7)            
      }
    }
  }  
    
  return tDefaults
end

function AlpacaProcs:PluginForm()
  local fnCreatePositionOptions = function () 
    return {
      { strLabel = "Left",      nValue = AlpacaWindowLibrary.EnumPosition.Left },
      { strLabel = "Right",     nValue = AlpacaWindowLibrary.EnumPosition.Right },
      { strLabel = "Top",       nValue = AlpacaWindowLibrary.EnumPosition.Top },
      { strLabel = "Bottom",    nValue = AlpacaWindowLibrary.EnumPosition.Bottom }
    }
  end
  
  local fnCreateDirectionOptions = function () 
    return {
      { strLabel = "Left",    nValue = AlpacaWindowLibrary.EnumDirection.Left },
      { strLabel = "Right",   nValue = AlpacaWindowLibrary.EnumDirection.Right },
      { strLabel = "Up",      nValue = AlpacaWindowLibrary.EnumDirection.Up },
      { strLabel = "Down",    nValue = AlpacaWindowLibrary.EnumDirection.Down }
    }
  end
    
  local arInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Procs"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Spacing",                 { "nSpacing" }, 0, 200, 1),
    
    AlpacaPluginLibrary:CreateHeader("Abilities"),
    AlpacaPluginLibrary:CreateSlider("Size (Width)",          { "tProcBar", "tSize", "nWidth" }, 10, 1000, 1),
    AlpacaPluginLibrary:CreateSlider("Size (Height)",         { "tProcBar", "tSize", "nHeight" }, 10, 1000, 1),
    AlpacaPluginLibrary:CreateSelect("Fill Direction",        { "tProcBar", "eFillDirection" }, fnCreateDirectionOptions()),
    AlpacaPluginLibrary:CreateSlider("Marker (Width)",        { "tProcBar", "tMarker", "nWidth" }, 0, 200, 1),
    AlpacaPluginLibrary:CreateSlider("Marker (Height)",       { "tProcBar", "tMarker", "nHeight" }, 0, 200, 1),
    AlpacaPluginLibrary:CreateColor("Marker Color",           { "tProcBar", "tMarker", "crColor" }),
    AlpacaPluginLibrary:CreateColor("Cooldown Color",         { "tProcBar", "crCooldown" }),
    AlpacaPluginLibrary:CreateSlider("Icon Size",             { "tProcBar", "nIconSize" }, 0, 200, 1),
    AlpacaPluginLibrary:CreateSelect("Icon Position",         { "tProcBar", "eIconPosition" }, fnCreatePositionOptions()),
    
    AlpacaPluginLibrary:CreateSubHeader("Engineer"),
    AlpacaPluginLibrary:CreateColor("Quick Burst", { "tProcBar", "tAbilityColors", "Quick Burst" }),
    AlpacaPluginLibrary:CreateColor("Feedback", { "tProcBar", "tAbilityColors", "Feedback" }),
    
    AlpacaPluginLibrary:CreateSubHeader("Medic"),
    AlpacaPluginLibrary:CreateColor("Atomize", { "tProcBar", "tAbilityColors", "Atomize" }),
    AlpacaPluginLibrary:CreateColor("Dual Shock", { "tProcBar", "tAbilityColors", "Dual Shock" }),
    
    AlpacaPluginLibrary:CreateSubHeader("Spellslinger"),
    AlpacaPluginLibrary:CreateColor("Flame Burst", { "tProcBar", "tAbilityColors", "Flame Burst" }),
    
    AlpacaPluginLibrary:CreateSubHeader("Stalker"),
    AlpacaPluginLibrary:CreateColor("Punish", { "tProcBar", "tAbilityColors", "Punish" }),
    AlpacaPluginLibrary:CreateColor("Decimate", { "tProcBar", "tAbilityColors", "Decimate" }),
    
    AlpacaPluginLibrary:CreateSubHeader("Engineer"),
    AlpacaPluginLibrary:CreateColor("Breaching Strikes", { "tProcBar", "tAbilityColors", "Breaching Strikes" }),
    AlpacaPluginLibrary:CreateColor("Atomic Spear", { "tProcBar", "tAbilityColors", "Atomic Spear" }),
  }
  
  return AlpacaPluginLibrary:CreateForm(arInputDefinitions)
end


function AlpacaProcs:OnPluginRegistered(wndDisplay)
  self.wndDisplay = wndDisplay
end

function AlpacaProcs:OnPluginSettingsChanged(tSettings)
  self.tSettings = tSettings
  self:RequestRebuild(0)
end

-- Addon

function AlpacaProcs:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaProcs:Init() 
  local bHasConfigureFunction = false
  local strConfigureButtonText = nil
  local tDependencies = { "AlpacaGameLibrary", "AlpacaPluginLibrary", "AlpacaStandardLibrary", "AlpacaWindowLibrary" }
  
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function AlpacaProcs:OnLoad()
  Apollo.LoadSprites("AlpacaSprites.xml")

  AlpacaGameLibrary = Apollo.GetPackage("AlpacaGameLibrary").tPackage
  AlpacaPluginLibrary = Apollo.GetPackage("AlpacaPluginLibrary").tPackage
  AlpacaStandardLibrary = Apollo.GetPackage("AlpacaStandardLibrary").tPackage
  AlpacaWindowLibrary = Apollo.GetPackage("AlpacaWindowLibrary").tPackage

  Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)
  
  -- events
  Apollo.RegisterEventHandler("AttackMissed", "OnAttackMissed", self)  
  Apollo.RegisterEventHandler("DamageOrHealingDone", "OnDamageOrHealingDone", self)  
  Apollo.RegisterEventHandler("PlayerChanged", "OnPlayerChanged", self)
  Apollo.RegisterEventHandler("CharacterCreated", "OnCharacterCreated", self)
  Apollo.RegisterEventHandler("PlayerLevelChange", "OnPlayerLevelChange", self)
  Apollo.RegisterEventHandler("AbilityBookChange", "OnAbilityBookChange", self)
  Apollo.RegisterEventHandler("SpecChanged", "OnSpecChanged", self)
  
  -- register
  AlpacaPluginLibrary:RegisterPlugin(self, "Procs")
end

-- Events

function AlpacaProcs:OnAttackMissed(unitCaster, unitTarget, eMissType)
  if unitTarget and unitTarget:IsThePlayer() then
    if eMissType == GameLib.CodeEnumMissType.Dodge then
      self.fLastDodge = os.clock()
    end
  end
end

function AlpacaProcs:OnDamageOrHealingDone(unitCaster, unitTarget, eDamageType, nDamage, nShieldDamaged, nAbsorptionAmount, bCritical) 
  if bCritical and unitCaster and unitCaster:IsThePlayer() then  
    local bIsHeal = eDamageType == GameLib.CodeEnumDamageType.Heal or eDamageType == GameLib.CodeEnumDamageType.HealShields
    
    if not bIsHeal or unitCaster:GetClassId() == GameLib.CodeEnumClass.Medic then    
      self.fLastCritical = os.clock()
    end
  end
end

function AlpacaProcs:OnPlayerChanged()
  self:RequestRebuild(1)
end

function AlpacaProcs:OnCharacterCreated()
  self:RequestRebuild(1)
end

function AlpacaProcs:OnPlayerLevelChange()
  self:RequestRebuild(1)
end

function AlpacaProcs:OnAbilityBookChange()
  self:RequestRebuild(1)
end

function AlpacaProcs:OnSpecChanged()
  self:RequestRebuild(1)
end

-- Timers

function AlpacaProcs:RequestRebuild(fDelay)
  self.fRebuildTime = os.clock() + (fDelay or 0)
end

function AlpacaProcs:IsReady() 
  if not self.bReady then
    if not AlpacaWindowLibrary:IsLoaded() then 
      return false
    end

    if not self.wndDisplay then
      return false
    end

    if not self.tSettings then 
      return false
    end

    self.bReady = true
  end
  
  return self.bReady
end

function AlpacaProcs:OnFrame()  
  if not self:IsReady() then
    return
  end
  
  self.unitPlayer = AlpacaGameLibrary:GetPlayerUnit()

  if not self.unitPlayer then
    return
  end
     
  local bRebuild = self.fRebuildTime and os.clock() >= self.fRebuildTime 

  if not self.wndProcBars or bRebuild then
    self:CreateProcBars()
  end

  if (bRebuild) then
    self.fRebuildTime = nil
  end
  
  self:RefreshProcBars()
end

-- ProcBars

function AlpacaProcs:CreateProcBars()
  if self.wndProcBars then 
    self.wndProcBars:Destroy() 
    self.wndProcBars = nil
  end
  
  if not self.tSettings.bEnable then return end
  
  local arAbilities = AlpacaGameLibrary:GetActionBarProcs()
  local wndProcBars = AlpacaWindowLibrary:CreateEmptyFrame(self.wndDisplay)  
    
  for _, tAbility in ipairs(arAbilities) do
    self:CreateProcBar(tAbility, wndProcBars)
  end

  self.wndProcBars = wndProcBars  
end

function AlpacaProcs:RefreshProcBars() 
  if not self.wndProcBars then return end

  local nHorizontal = self.tSettings.tPosition.nHorizontal
  local nVertical = self.tSettings.tPosition.nVertical
  local eDirection = self.tSettings.tProcBar.nFillDirection
  local nWidth = self.tSettings.tProcBar.tSize.nWidth
  local nHeight = self.tSettings.tProcBar.tSize.nHeight
  local nSpacing = self.tSettings.nSpacing
  
  -- refresh bars

  local nVisible = 0

  for _, wndProcBar in ipairs(self.wndProcBars:GetChildren()) do
    if not  self:RefreshProcBar(wndProcBar) then
      break
    end
    
    local nOffsetLeft = 0
    local nOffsetTop = 0
    
    if eDirection == AlpacaWindowLibrary.EnumDirection.Left or eDirection == AlpacaWindowLibrary.EnumDirection.Right then
      nOffsetLeft = 0
      nOffsetTop = nVisible * (nHeight + nSpacing)
    else
      nOffsetLeft = nVisible * (nWidth + nSpacing)
      nOffsetTop = 0
    end

    wndProcBar:SetAnchorOffsets(nOffsetLeft, nOffsetTop, nOffsetLeft + nWidth, nOffsetTop + nHeight)

    nVisible = nVisible + 1
  end
  
  -- resize container
  
  if eDirection == AlpacaWindowLibrary.EnumDirection.Left or eDirection == AlpacaWindowLibrary.EnumDirection.Right then
    nHeight = nHeight * nVisible + nSpacing * (nVisible - 1) 
  else
    nWidth = nWidth * nVisible + nSpacing * (nVisible - 1)
  end
  
  local nLeft = nHorizontal - math.ceil(nWidth / 2)
  local nTop = nVertical - math.ceil(nHeight / 2)
  local nRight = nLeft + nWidth
  local nBottom = nTop + nHeight

  self.wndProcBars:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
  
  -- toggle frame
    
  local bShow = true

  if self.tSettings.bOnlyShowInCombat and not AlpacaGameLibrary:IsInCombatOrDamaged(self.unitPlayer) then
    bShow = false
  end
  
  if self.tSettings.bPreview then
    bShow = true
  end

  self.wndProcBars:Show(bShow, true)
end

-- ProcBar

function AlpacaProcs:CreateProcBar(tAbility, wndParent)
  local wndBar = AlpacaWindowLibrary:CreateBar(self.tSettings.tProcBar, wndParent)   
  local tData = { tAbility = tAbility }
  
  wndBar:SetData(tData)
  
  local wndBorder = wndBar:FindChild("Border")
  local crBorder = self.tSettings.tProcBar.tAbilityColors[tAbility.strName]
  
  wndBorder:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crBorder))
  
  local wndIcon = wndBar:FindChild("Icon")  
  local strIcon = tAbility.tTiers[tAbility.nCurrentTier].splObject:GetIcon()
  
  if wndIcon then
    wndIcon:SetSprite(strIcon)
  end
  
  local wndProgressBar = wndBar:FindChild("ProgressBar")
  
  AlpacaWindowLibrary:CreateMarker(AlpacaWindowLibrary.EnumMarkerType.Arrow, self.tSettings.tProcBar, wndProgressBar, 0)
end

function AlpacaProcs:RefreshProcBar(wndProcBar) 
  local tData = wndProcBar:GetData() 
  
  -- get proc times
  
  local fProcRemaining, fProcDuration, fCooldownRemaining = AlpacaGameLibrary:GetProcTimes(tData.tAbility, self.fLastCritical, self.fLastDodge)
  
  if self.tSettings.bPreview then
    fProcRemaining = 3
    fCooldownRemaining = 1
  end

  -- hide if proc expired or ability will not be off cooldown before proc expiration
  
  if not fProcRemaining or (fCooldownRemaining and fCooldownRemaining >= fProcRemaining) then
    wndProcBar:Show(false)
    return false
  end
  
  -- update proc time left
  
  local fPercent = math.max(0, math.min(1, fProcRemaining / fProcDuration))
  AlpacaWindowLibrary:UpdateBarFill(wndProcBar, self.tSettings.tProcBar, fPercent)
  
  -- update color
  
  local crColor = nil 
  
  if fCooldownRemaining then
    crColor = self.tSettings.tProcBar.crCooldown
  else
    crColor = self.tSettings.tProcBar.tAbilityColors[tData.tAbility.strName]
  end
  
  local wndFill = wndProcBar:FindChild("Fill")
  wndFill:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crColor))
  
  -- update marker
  
  local wndMarker = wndProcBar:FindChild("Marker")

  if fCooldownRemaining then
    AlpacaWindowLibrary:MoveMarker(wndMarker, self.tSettings.tProcBar, (fProcRemaining - fCooldownRemaining) / fProcDuration)
  end
  
  wndMarker:Show(fCooldownRemaining)    
  
  -- show
  
  wndProcBar:Show(true)  
  
  return true
end

-- Start

local AlpacaProcsInst = AlpacaProcs:new()
AlpacaProcsInst:Init()
